import typing
import logging
import inspect
from dataclasses import MISSING

logger = logging.getLogger(__name__)


def get_function_call_args(
    func, params: typing.Union[typing.Dict[str, typing.Any], object]
) -> typing.Dict[str, typing.Any]:
    """
    Extracts the arguments for a function call from the provided parameters.

    Args:
        func: The function for which arguments are to be extracted.
        params: A dictionary of parameters containing
                the necessary arguments for the function.

    Returns:
        A dictionary where the keys are the function argument names
        and the values are the corresponding argument values.
    """
    params_dict = {}
    try:
        sig = inspect.signature(func)
        for param in sig.parameters.values():
            if param.name != "self":
                value = (
                    params.get(param.name, param.default)
                    if isinstance(params, dict)
                    else getattr(params, param.name, param.default)
                )
                if value is not MISSING and value is not inspect.Parameter.empty:
                    params_dict[param.name] = value
                else:
                    params_dict[param.name] = None
    except (ValueError, KeyError) as e:
        logger.warning(f"Parsing {func} for call parameters failed {str(e)}")

    for key in ["args", "kwargs"]:
        if key in params_dict and params_dict[key] is None:
            params_dict.pop(key, None)
    return params_dict


def init_class(
    klass: typing.Type, params: typing.Union[typing.Dict[str, typing.Any], object]
):
    kwargs = get_function_call_args(klass.__init__, params)
    excluded_kwargs = {key: params[key] for key in params if key not in kwargs}
    instance = klass(**kwargs)
    if excluded_kwargs:
        instance.__dict__.update(excluded_kwargs)
    return instance



import csv
import json
import typing
from dataclasses import asdict
from abc import ABC, abstractmethod

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

from .utils import init_class

if typing.TYPE_CHECKING:
    from .base import BaseModel


_BLOCK_SIZE = 1024

T = typing.TypeVar("T", typing.List["BaseModel"], "BaseModel")
D = typing.TypeVar(
    "D", typing.Dict[str, typing.Any], typing.List[typing.Dict[str, typing.Any]]
)


class BaseModelFormatter(ABC):
    format_name: str = None

    @classmethod
    def is_format_name(cls, format_name: str) -> bool:
        format_names = (
            isinstance(cls.format_name, (list, tuple))
            and format_name
            or [cls.format_name]
        )
        return format_name in format_names

    @abstractmethod
    def encode(self, _type: typing.Type["BaseModel"], obj: D) -> T:
        pass

    @abstractmethod
    def decode(self, instance: "BaseModel") -> typing.Any:
        pass

    @classmethod
    def get_formatters(cls):
        for subclass in cls.__subclasses__():
            yield from subclass.get_formatters()
            yield subclass

    @classmethod
    def get_formatter(cls, format_name: str, **config) -> "BaseModelFormatter":
        for subclass in cls.get_formatters():
            if subclass.is_format_name(format_name):
                return subclass(**config)
        raise KeyError(f"Format {format_name} not found")


class DictModelFormatter(BaseModelFormatter):
    format_name = "dict"

    def _encode(
        self, _type: typing.Type["BaseModel"], obj: typing.Dict[str, typing.Any]
    ) -> "BaseModel":
        instance = init_class(_type, obj)
        # force execute post init again for normal field validation
        instance.__post_init__()
        return instance

    def encode(self, _type: typing.Type["BaseModel"], obj: D) -> T:
        if isinstance(obj, dict):
            return self._encode(_type, obj)
        elif isinstance(obj, list):
            return [self._encode(_type, item) for item in obj]
        else:
            raise TypeError("Object must be dict or list")

    def decode(self, instance: T) -> D:
        if isinstance(instance, list):
            return [asdict(val) for val in instance]
        return asdict(instance)


class JSONModelFormatter(DictModelFormatter):
    format_name = "json"

    def encode(self, _type: typing.Type["BaseModel"], obj: str) -> T:
        return super().encode(_type, json.loads(obj))

    def decode(self, instance: T) -> str:
        return json.dumps(super().decode(instance))


class CSVModelFormatter(DictModelFormatter):
    format_name = "csv"

    def encode(self, _type: typing.Type["BaseModel"], file: str) -> T:
        with open(file, "r", newline="") as f:
            sample = f.read(_BLOCK_SIZE)
            dialect = csv.Sniffer().sniff(sample)
            has_header = csv.Sniffer().has_header(sample)
            f.seek(0)
            if not has_header:
                raise FileExistsError(f"File {file} does not have header")
            reader = csv.DictReader(f, dialect=dialect)
            return [super().encode(_type, row) for row in reader]

    def decode(self, instance: T) -> str:
        instances = instance if isinstance(instance, (list, tuple)) else [instance]
        with StringIO() as f:
            writer = csv.DictWriter(f, dialect=csv.excel, fieldnames=[])
            for index, obj in enumerate(instances):
                instance_dict = super().decode(obj)
                if index == 0:
                    writer.fieldnames = list(instance_dict.keys())
                    writer.writeheader()
                writer.writerow(instance_dict)

            context = f.getvalue()

        return context
